<?php

namespace App\Domain\Catalogs\Models;

use App\Domain\Catalogs\Models\Tests\Factories\CatalogItemsFactory;
use App\Domain\Categories\Models\Category;
use App\Domain\Support\Models\Model;
use Cviebrock\EloquentSluggable\Sluggable;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Database\Eloquent\Relations\HasMany;
use Illuminate\Support\Collection;
use Kalnoy\Nestedset\NodeTrait;

/**
 * Class-model for entity "Catalog Item".
 *
 * @property string $name
 * @property string $code Catalog item unique code
 * @property string|null $url Page url
 * @property int|null $parent_id
 * @property int|null $category_id
 * @property int $catalog_id
 * @property bool $is_active
 *
 * @property-read CatalogItem|null $parent Parent catalog item
 * @property-read Collection|CatalogItem[] $ancestors Parent catalog items
 * @property-read Collection|CatalogItem[] $children Child catalog items
 * @property-read Collection|CatalogItem[] $descendants Child catalog items of all levels
 *
 * @property-read Category|null $category Category in catalog item
 * @property-read Catalog $catalog
 *
 * @method static Collection|CatalogItem[] ancestorsAndSelf(int $id)
 * @method static Collection|CatalogItem[] descendantsAndSelf(int $id, array $columns = ['*'])
 * @method static Builder|self whereAncestorOf(int $catalogItemId, bool $includeSelf)
 * @method static Builder|self whereIsRoot()
 */
class CatalogItem extends Model
{
    use NodeTrait;
    use Sluggable;

    protected $table = 'catalog_items';
    protected $fillable = ['name', 'code', 'url', 'parent_id', 'catalog_id', 'category_id', 'is_active'];

    protected $attributes = [
        'is_active' => false,
    ];

    protected $casts = [
        'parent_id' => 'int',
        'is_active' => 'bool',
    ];

    protected $hidden = ['_lft', '_rgt'];

    public function catalog(): BelongsTo
    {
        return $this->belongsTo(Catalog::class, 'catalog_id');
    }

    public function category(): BelongsTo
    {
        return $this->belongsTo(Category::class, 'category_id');
    }

    public function children(): HasMany
    {
        return $this->hasMany(CatalogItem::class, 'parent_id')->with('children');
    }

    public function sluggable(): array
    {
        return [
            'code' => [
                'source' => 'name',
                'separator' => '_',
                'unique' => true,
                'maxLength' => 30,
            ],
        ];
    }

    public function replicate(array $except = null): self
    {
        $defaults = [
            $this->getParentIdName(),
            $this->getLftName(),
            $this->getRgtName(),
            'code',
        ];

        $except = empty($except) ? $defaults : array_unique(array_merge($except, $defaults));

        return parent::replicate($except);
    }

    public function scopeForParent(Builder $query, int $parentId): Builder
    {
        return $query->where($this->getParentIdName(), $parentId);
    }

    public static function factory(): CatalogItemsFactory
    {
        return CatalogItemsFactory::new();
    }
}
