<?php

namespace App\Http\ApiV1\Modules\Catalogs\Resources;

use App\Domain\Catalogs\Models\CatalogItem;
use Illuminate\Http\Resources\Json\ResourceCollection;
use Illuminate\Support\Collection;

class CatalogMultiResourceCollection extends ResourceCollection
{
    protected bool $onlyActive;

    public function __construct(Collection $resource, bool $onlyActive)
    {
        $this->onlyActive = $onlyActive;
        parent::__construct($resource);
    }

    public function toArray($request): array
    {
        return $this->collection->map(fn ($catalog) => [
            'id' => $catalog->id,
            'code' => $catalog->code,
            'name' => $catalog->name,
            'is_active' => $catalog->is_active,
            'sections' => $this->buildTree($catalog->catalogItems),
        ])->toArray();
    }

    private function buildTree(Collection $items): array
    {
        $items = $this->onlyActive
            ? $items->where('is_active', true)
            : $items;

        $itemsById = $items->keyBy('id');
        $tree = [];

        foreach ($items as $item) {
            if ($item->parent_id && $itemsById->has($item->parent_id)) {
                $parent = $itemsById->get($item->parent_id);

                if (!$parent->relationLoaded('children')) {
                    $parent->setRelation('children', collect());
                }

                $parent->children->push($item);
            } else {
                $tree[] = $this->catalogItemToArray($item);
            }
        }

        return array_values(array_filter($tree));
    }

    private function catalogItemToArray(CatalogItem $item): ?array
    {
        if ($this->onlyActive && !$item->is_active) {
            return null;
        }

        $children = $item->relationLoaded('children')
            ? $this->mapItems($item->children)
            : [];

        return [
            'id' => $item->id,
            'name' => $item->name,
            'code' => $item->code,
            'url' => $item->url,
            'category_id' => $item->category_id,
            'is_active' => $item->is_active,
            'children' => $children,
        ];
    }

    private function mapItems(Collection $items): array
    {
        return $items
            ->filter(fn (CatalogItem $item) => !$this->onlyActive || $item->is_active)
            ->map(fn (CatalogItem $item) => $this->catalogItemToArray($item))
            ->values()
            ->toArray();
    }
}
