<?php

use App\Domain\Catalogs\Models\Catalog;
use App\Domain\Catalogs\Models\CatalogItem;
use App\Http\ApiV1\Modules\Catalogs\Tests\Factories\CatalogItemRequestFactory;
use App\Http\ApiV1\OpenApiGenerated\Enums\PaginationTypeEnum;
use App\Http\ApiV1\Support\Tests\ApiV1ComponentTestCase;

use function Pest\Laravel\assertDatabaseHas;
use function Pest\Laravel\assertModelMissing;
use function Pest\Laravel\deleteJson;
use function Pest\Laravel\getJson;
use function Pest\Laravel\patchJson;
use function Pest\Laravel\postJson;
use function Pest\Laravel\putJson;

uses(ApiV1ComponentTestCase::class);
uses()->group('component', 'catalogs');

test('POST /api/v1/catalogs/catalog-items success', function () {
    $request = CatalogItemRequestFactory::new()->forCatalog()->forCategory()->make();

    postJson('/api/v1/catalogs/catalog-items', $request)
        ->assertStatus(201);

    assertDatabaseHas(CatalogItem::class, ['code' => $request['code']]);
});

test('POST /api/v1/catalogs/catalog-items 400 item already exists', function () {
    $catalogItem = CatalogItem::factory()->create();
    $request = CatalogItemRequestFactory::new()->forCatalog()->make(['code' => $catalogItem->code]);

    postJson('/api/v1/catalogs/catalog-items', $request)
        ->assertStatus(400);
});

test('POST /api/v1/catalogs/catalog-items 400 value does not exist', function (?string $key, $value) {
    $request = CatalogItemRequestFactory::new()->make([$key => $value]);

    postJson('/api/v1/catalogs/catalog-items', $request)
        ->assertStatus(400);
})->with([
    ['catalog_id', 12],
    ['category_id', 12],
    ['parent_id', 12],
]);

test('POST /api/v1/catalogs/catalog-items 400 wrong catalog_id', function () {
    $catalog1 = Catalog::factory()->create();
    $catalog2 = Catalog::factory()->create();
    $catalogItem = CatalogItem::factory()->withCatalog($catalog1)->create();
    $request = CatalogItemRequestFactory::new()->forCatalog($catalog2)->forParent($catalogItem)->make();

    postJson('/api/v1/catalogs/catalog-items', $request)
        ->assertStatus(400);
});

test('GET /api/v1/catalogs/catalog-items/{id} 200', function () {
    $catalogItem = CatalogItem::factory()->create();

    getJson("/api/v1/catalogs/catalog-items/$catalogItem->id")
        ->assertStatus(200)
        ->assertJsonStructure(['data' => ['id', 'code', 'name']]);
});

test('PUT /api/v1/catalogs/catalog-items/{id} 200', function () {
    $catalogItem = CatalogItem::factory()->create();
    $id = $catalogItem->id;
    $request = CatalogItemRequestFactory::new()->forCatalog()->forCategory()->make();

    putJson("/api/v1/catalogs/catalog-items/$id", $request)
        ->assertStatus(200)
        ->assertJsonPath('data.code', $request['code']);

    assertDatabaseHas(CatalogItem::class, ['id' => $id, 'code' => $request['code'], 'name' => $request['name']]);
});

test('PUT /api/v1/catalogs/catalog-items/{id} 400 value does not exist', function (?string $key, $value) {
    $catalogItem = CatalogItem::factory()->create();
    $id = $catalogItem->id;
    $request = CatalogItemRequestFactory::new()->make([$key => $value]);

    putJson("/api/v1/catalogs/catalog-items/$id", $request)
        ->assertStatus(400);
})->with([
    ['catalog_id', 12],
    ['category_id', 12],
    ['parent_id', 12],
]);

test('DELETE /api/v1/catalogs/catalog-items/{id} 200', function () {
    $catalogItem = CatalogItem::factory()->create();
    $id = $catalogItem->id;

    deleteJson("/api/v1/catalogs/catalog-items/$id")
        ->assertStatus(200);

    assertModelMissing($catalogItem);
});

test('DELETE /api/v1/catalogs/catalog-items/{id} 400 parent category fails', function () {
    $catalogItem = CatalogItem::factory()->createTree()->first();
    $id = $catalogItem->id;

    deleteJson("/api/v1/catalogs/catalog-items/$id")
        ->assertStatus(400);

    assertDatabaseHas(CatalogItem::class, ['id' => $id]);
});

test('PATCH /api/v1/catalogs/catalog-items/{id} 200', function () {
    $catalogItem = CatalogItem::factory()->create();
    $id = $catalogItem->id;
    $request = CatalogItemRequestFactory::new()->only(['code'])->make();

    patchJson("/api/v1/catalogs/catalog-items/$id", $request)
        ->assertStatus(200)
        ->assertJsonPath('data.code', $request['code']);

    assertDatabaseHas(CatalogItem::class, ['id' => $id, 'code' => $request['code']]);
});

test('PATCH /api/v1/catalogs/catalog-items/{id} 400 catalog_id does not exist', function (?string $key, $value) {
    $catalogItem = CatalogItem::factory()->create();
    $id = $catalogItem->id;
    $request = CatalogItemRequestFactory::new()->only([$value])->make([$key => $value]);

    patchJson("/api/v1/catalogs/catalog-items/$id", $request)
        ->assertStatus(400);
})->with([
    ['catalog_id', 12],
    ['category_id', 12],
    ['parent_id', 12],
]);

test('POST /api/v1/catalogs/catalog-items:search 200', function () {
    $catalogItem = CatalogItem::factory()->create();
    $request = [
        'filter' => ['code' => $catalogItem->code],
    ];

    postJson('/api/v1/catalogs/catalog-items:search', $request)
        ->assertStatus(200)
        ->assertJsonCount(1, 'data')
        ->assertJsonPath('data.0.code', $catalogItem->code);
});

test('POST /api/v1/catalogs/catalog-items:search sort success', function (string $sort) {
    $catalogItem = CatalogItem::factory()->create();

    postJson('/api/v1/catalogs/catalog-items:search', ["sort" => [$sort]])
        ->assertStatus(200)
        ->assertJsonCount(1, 'data')
        ->assertJsonPath('data.0.code', $catalogItem->code);
})->with([
    'id',
    'code',
    'name',
    'url',
    'is_active',
    'updated_at',
    'created_at',
]);

test('POST /api/v1/catalogs/catalog-items:search filter success', function (string $fieldKey, $value = null, ?string $filterKey = null, $filterValue = null) {
    /** @var CatalogItem $model */
    $model = CatalogItem::factory()->create($value ? [$fieldKey => $value] : []);

    postJson('/api/v1/catalogs/catalog-items:search', ["filter" => [
        ($filterKey ?: $fieldKey) => ($filterValue ?: $model->{$fieldKey}),
    ], 'sort' => ['id'], 'pagination' => ['type' => PaginationTypeEnum::CURSOR, 'limit' => 1]])
        ->assertStatus(200)
        ->assertJsonCount(1, 'data')
        ->assertJsonPath('data.0.id', $model->id);
})->with([
    ['id'],
    ['name'],
    ['code'],
    ['url'],
    ['parent_id'],
    ['catalog_id'],
    ['is_active'],
    ['created_at', '2022-04-20T01:32:08.000000Z', 'created_at_gte', '2022-04-19T01:32:08.000000Z'],
    ['created_at', '2022-04-20T01:32:08.000000Z', 'created_at_lte', '2022-04-21T01:32:08.000000Z'],
    ['updated_at', '2022-04-20T01:32:08.000000Z', 'updated_at_gte', '2022-04-19T01:32:08.000000Z'],
    ['updated_at', '2022-04-20T01:32:08.000000Z', 'updated_at_lte', '2022-04-21T01:32:08.000000Z'],
]);

test('POST /api/v1/catalogs/catalog-items:search include success', function () {
    /** @var CatalogItem $catalogItem */
    $catalogItem = CatalogItem::factory()
        ->withCategory()
        ->withCatalog()
        ->create();

    postJson('/api/v1/catalogs/catalog-items:search', ["include" => ['catalog', 'category']])
        ->assertStatus(200)
        ->assertJsonCount(1, 'data')
        ->assertJsonPath('data.0.category.id', $catalogItem->category->id)
        ->assertJsonPath('data.0.catalog.id', $catalogItem->catalog->id);
});
