<?php

use App\Domain\Catalogs\Models\Catalog;
use App\Domain\Catalogs\Models\CatalogItem;
use App\Http\ApiV1\Modules\Catalogs\Tests\Factories\CatalogRequestFactory;
use App\Http\ApiV1\OpenApiGenerated\Enums\PaginationTypeEnum;
use App\Http\ApiV1\Support\Tests\ApiV1ComponentTestCase;

use function Pest\Laravel\assertDatabaseHas;
use function Pest\Laravel\assertModelMissing;
use function Pest\Laravel\deleteJson;
use function Pest\Laravel\getJson;
use function Pest\Laravel\patchJson;
use function Pest\Laravel\postJson;
use function Pest\Laravel\putJson;

uses(ApiV1ComponentTestCase::class);
uses()->group('component', 'catalogs');

test('POST /api/v1/catalogs/catalogs success', function () {
    $request = CatalogRequestFactory::new()->make();

    postJson('/api/v1/catalogs/catalogs', $request)
        ->assertStatus(201);

    assertDatabaseHas(Catalog::class, ['code' => $request['code']]);
});

test('POST /api/v1/catalogs/catalogs 400 not unique code', function () {
    $catalog = Catalog::factory()->create();
    $request = CatalogRequestFactory::new()->make(['code' => $catalog->code]);

    postJson('/api/v1/catalogs/catalogs', $request)
        ->assertStatus(400);
});

test('GET /api/v1/catalogs/catalogs/{id} 200', function () {
    $catalog = Catalog::factory()->create();

    getJson("/api/v1/catalogs/catalogs/$catalog->id")
        ->assertStatus(200)
        ->assertJsonStructure(['data' => ['id', 'code', 'name']]);
});

test('PUT /api/v1/catalogs/catalogs/{id} 200', function () {
    $catalog = Catalog::factory()->create();
    $id = $catalog->id;
    $request = CatalogRequestFactory::new()->make();

    putJson("/api/v1/catalogs/catalogs/$id", $request)
        ->assertStatus(200)
        ->assertJsonPath('data.code', $request['code']);

    assertDatabaseHas(Catalog::class, ['id' => $id, 'code' => $request['code'], 'name' => $request['name']]);
});

test('DELETE /api/v1/catalogs/catalogs/{id} 200', function () {
    $catalog = Catalog::factory()->create();
    $id = $catalog->id;

    deleteJson("/api/v1/catalogs/catalogs/$id")
        ->assertStatus(200);

    assertModelMissing($catalog);
});

test('DELETE /api/v1/catalogs/catalogs/{id} 400 items attached', function () {
    $catalog = Catalog::factory()
        ->has(CatalogItem::factory()->count(3), 'catalogItems')
        ->create();
    $id = $catalog->id;

    deleteJson("/api/v1/catalogs/catalogs/$id")
        ->assertStatus(400);
});

test('PATCH /api/v1/catalogs/catalogs/{id} 200', function () {
    $catalog = Catalog::factory()->create();
    $id = $catalog->id;
    $request = CatalogRequestFactory::new()->only(['code'])->make();

    patchJson("/api/v1/catalogs/catalogs/$id", $request)
        ->assertStatus(200)
        ->assertJsonPath('data.code', $request['code']);

    assertDatabaseHas(Catalog::class, ['id' => $id, 'code' => $request['code']]);
});

test('POST /api/v1/catalogs/catalogs:search 200', function () {
    $catalog = Catalog::factory()->create();
    $request = [
        'filter' => ['code' => $catalog->code],
    ];

    postJson('/api/v1/catalogs/catalogs:search', $request)
        ->assertStatus(200)
        ->assertJsonCount(1, 'data')
        ->assertJsonPath('data.0.code', $catalog->code);
});

test('POST /api/v1/catalogs/catalogs:search sort success', function (string $sort) {
    $catalog = Catalog::factory()->create();

    postJson('/api/v1/catalogs/catalogs:search', ["sort" => [$sort]])
        ->assertStatus(200)
        ->assertJsonCount(1, 'data')
        ->assertJsonPath('data.0.code', $catalog->code);
})->with([
    'id',
    'code',
    'name',
    'is_active',
    'updated_at',
    'created_at',
]);

test('POST /api/v1/catalogs/catalogs:search filter success', function (string $fieldKey, $value = null, ?string $filterKey = null, $filterValue = null) {
    /** @var Catalog $model */
    $model = Catalog::factory()->create($value ? [$fieldKey => $value] : []);

    postJson('/api/v1/catalogs/catalogs:search', ["filter" => [
        ($filterKey ?: $fieldKey) => ($filterValue ?: $model->{$fieldKey}),
    ], 'sort' => ['id'], 'pagination' => ['type' => PaginationTypeEnum::CURSOR, 'limit' => 1]])
        ->assertStatus(200)
        ->assertJsonCount(1, 'data')
        ->assertJsonPath('data.0.id', $model->id);
})->with([
    ['id'],
    ['name'],
    ['code'],
    ['is_active'],
    ['created_at', '2022-04-20T01:32:08.000000Z', 'created_at_gte', '2022-04-19T01:32:08.000000Z'],
    ['created_at', '2022-04-20T01:32:08.000000Z', 'created_at_lte', '2022-04-21T01:32:08.000000Z'],
    ['updated_at', '2022-04-20T01:32:08.000000Z', 'updated_at_gte', '2022-04-19T01:32:08.000000Z'],
    ['updated_at', '2022-04-20T01:32:08.000000Z', 'updated_at_lte', '2022-04-21T01:32:08.000000Z'],
]);

test('POST /api/v1/catalogs/catalogs:search include success', function () {
    Catalog::factory()
        ->withCatalogItems()
        ->create();

    postJson('/api/v1/catalogs/catalogs:search', ["include" => ['catalog_items']])
        ->assertStatus(200)
        ->assertJsonCount(1, 'data')
        ->assertJsonCount(3, 'data.0.catalog_items');
});

test('POST /api/v1/catalogs/multi-catalog in depth success', function () {
    $catalog = Catalog::factory()->create(['is_active' => true]);
    $catalogItems = CatalogItem::factory()->withCatalog($catalog)->createTree(3);
    $deepest = $catalogItems->last();

    postJson("/api/v1/catalogs/multi-catalog", ["include" => ['sections']])
        ->assertStatus(200)
        ->assertJsonStructure([
            'data' => [
                [
                    'id',
                    'code',
                    'name',
                    'sections' => [
                        '*' => [
                            'id',
                            'name',
                            'code',
                            'url',
                            'category_id',
                            'children',
                        ],
                    ],
                ],
            ],
        ])
        ->assertJsonFragment(['id' => $deepest->id, 'name' => $deepest->name]);
});

test('POST /api/v1/catalogs/multi-catalog across success', function () {
    $count = 2;
    $catalog = Catalog::factory()->create(['is_active' => true]);
    CatalogItem::factory()->withCatalog($catalog)->count($count)->create(['is_active' => true]);

    postJson("/api/v1/catalogs/multi-catalog", ["include" => ['sections']])
        ->assertStatus(200)
        ->assertJsonCount($count, 'data.0.sections');
});

test('POST /api/v1/catalogs/multi-catalog only active success', function () {
    $catalog = Catalog::factory()->create(['is_active' => true]);
    $root = CatalogItem::factory()->withCatalog($catalog)->create(['is_active' => true]);
    $parent = CatalogItem::factory()->withCatalog($catalog)->withParent($root)->create(['is_active' => false]);
    $child = CatalogItem::factory()->withCatalog($catalog)->withParent($parent)->create(['is_active' => false]);

    postJson("/api/v1/catalogs/multi-catalog", ["include" => ['sections'], "filter" => ['is_active' => true]])
        ->assertStatus(200)
        ->assertJsonCount(1, 'data')
        ->assertJsonCount(0, 'data.0.sections.0.children')
        ->assertJsonMissing(['id' => $child->id]);
});
